@@ -1,36 +1,38 @@
 #  ExoIris: fast, flexible, and easy exoplanet transmission spectroscopy in Python.
 #  Copyright (C) 2024 Hannu Parviainen
 #
-#   This program is free software: you can redistribute it and/or modify
-#   it under the terms of the GNU General Public License as published by
-#   the Free Software Foundation, either version 3 of the License, or
-#   (at your option) any later version.
+#  This program is free software: you can redistribute it and/or modify
+#  it under the terms of the GNU General Public License as published by
+#  the Free Software Foundation, either version 3 of the License, or
+#  (at your option) any later version.
 #
-#   This program is distributed in the hope that it will be useful,
-#   but WITHOUT ANY WARRANTY; without even the implied warranty of
-#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-#   GNU General Public License for more details.
+#  This program is distributed in the hope that it will be useful,
+#  but WITHOUT ANY WARRANTY; without even the implied warranty of
+#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+#  GNU General Public License for more details.
 #
-#   You should have received a copy of the GNU General Public License
-#   along with this program.  If not, see <https://www.gnu.org/licenses/>.
-from copy import deepcopy
-from typing import Optional
+#  You should have received a copy of the GNU General Public License
+#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
-from ldtk import BoxcarFilter, LDPSetCreator
+from copy import deepcopy
+from typing import Optional, Literal
+
+from ldtk import BoxcarFilter, LDPSetCreator   # noqa
 from numba import njit, prange
 from numpy import zeros, log, pi, linspace, inf, atleast_2d, newaxis, clip, arctan2, ones, floor, sum, concatenate, \
-    sort, ndarray, zeros_like, array, tile, arange, squeeze
+    sort, ndarray, zeros_like, array, tile, arange, squeeze, dstack
 from numpy.random import default_rng
 from celerite2 import GaussianProcess as GP, terms
 
-from pytransit import TSModel, RRModel, LDTkLD, TransitAnalysis
 from pytransit.lpf.logposteriorfunction import LogPosteriorFunction
 
 from pytransit.orbits import as_from_rhop, i_from_ba, fold, i_from_baew, d_from_pkaiews, epoch
 from pytransit.param import ParameterSet, UniformPrior as UP, NormalPrior as NP, GParameter
-from scipy.interpolate import splev, splrep
+from scipy.interpolate import pchip_interpolate, splrep, splev, Akima1DInterpolator
 
-from .tsdata import TSDataSet
+from .tsmodel import TransmissionSpectroscopyModel as TSModel
+from .tsdata import TSDataGroup
+from .ldtkld import LDTkLD
 
 NM_WHITE = 0
 NM_GP_FIXED = 1
@@ -38,20 +40,45 @@ NM_GP_FREE = 2
 
 noise_models = dict(white=NM_WHITE, fixed_gp=NM_GP_FIXED, free_gp=NM_GP_FREE)
 
+
 @njit(parallel=True, cache=False)
-def lnlike_normal(o, m, e):
+def lnlike_normal(o, m, e, f, mask):
+    nwl, nt = o.shape
+
     if m.ndim == 2:
-        return -sum(log(e)) - 0.5 * o.size * log(2. * pi) - 0.5 * sum((o - m) ** 2 / e ** 2)
+        la, lb, lc = 0.0, 0.0, 0.0
+        for iwl in range(nwl):
+            for it in range(nt):
+                if mask[iwl, it]:
+                    la += log(f[0]*e[iwl, it])
+                    lb += log(2*pi)
+                    lc += (o[iwl, it] - m[iwl, it])**2 / (f[0]*e[iwl, it])**2
+        return -la - 0.5*lb - 0.5*lc
     if m.ndim == 3:
         npv = m.shape[0]
         lnlike = zeros(npv)
         for i in prange(npv):
-            lnlike[i] = -sum(log(e)) - 0.5 * o.size * log(2. * pi) - 0.5 * sum((o - m[i]) ** 2 / e ** 2)
+            la, lb, lc = 0.0, 0.0, 0.0
+            for iwl in range(nwl):
+                for it in range(nt):
+                    if mask[iwl, it]:
+                        la += log(f[i]*e[iwl, it])
+                        lb += log(2*pi)
+                        lc += (o[iwl, it] - m[i, iwl, it])**2 / (f[i]*e[iwl, it])**2
+            lnlike[i] = -la - 0.5*lb - 0.5*lc
         return lnlike
 
 
-def resample(x_new, x_old, y_old):
-    return splev(x_new, splrep(x_old, y_old, s=0.0))
+def ip_pchip(x, xk, yk):
+    return pchip_interpolate(xk, yk, x)
+
+
+def ip_bspline(x, xk, yk):
+    return splev(x, splrep(xk, yk))
+
+
+def ip_makima(x, xk, yk):
+    return Akima1DInterpolator(xk, yk, method='makima', extrapolate=True)(x)
 
 
 def add_knots(x_new, x_old):
@@ -97,26 +124,33 @@ def clean_knots(knots, min_distance, lmin=0, lmax=inf):
 
 
 class TSLPF(LogPosteriorFunction):
-    def __init__(self, name: str, ldmodel, data: TSDataSet, nk: int = 50, nldc: int = 10, nthreads: int = 1,
-                 tmpars = None, noise_model: str = 'white'):
+    def __init__(self, runner, name: str, ldmodel, data: TSDataGroup, nk: int = 50, nldc: int = 10, nthreads: int = 1,
+                 tmpars = None, noise_model: Literal["white", "fixed_gp", "free_gp"] = 'white',
+                 interpolation: Literal['bspline', 'pchip', 'makima'] = 'bspline'):
         super().__init__(name)
-        self._original_data: TSDataSet | None = None
-        self.data: TSDataSet | None = None
+        self._runner = runner
+        self._original_data: TSDataGroup | None = None
+        self.data: TSDataGroup | None = None
         self.npb: list[int] | None= None
         self.npt: list[int] | None = None
         self.ndim: int | None = None
+        self._baseline_models: list[ndarray] | None = None
+        self.interpolation: str = interpolation
+
+        self._ip = {'bspline': ip_bspline, 'pchip': ip_pchip, 'makima': ip_makima}[interpolation]
 
         self._gp: Optional[list[GP]] = None
         self._gp_time: Optional[list[ndarray]] = None
         self._gp_flux: Optional[list[ndarray]] = None
 
+        self.tms = [TSModel(ldmodel, nthreads=nthreads, **(tmpars or {})) for i in range(len(data))]
+        self.set_data(data)
         self.set_noise_model(noise_model)
 
         self.ldmodel = ldmodel
-        self.tms = [TSModel(ldmodel, nthreads=nthreads, **(tmpars or {})) for i in range(len(data))]
-        self.set_data(data)
-
         if isinstance(ldmodel, LDTkLD):
+            for tm in self.tms:
+                tm.ldmodel = None
             self.ldmodel._init_interpolation(self.tms[0].mu)
 
         self.nthreads = nthreads
@@ -143,33 +177,37 @@ class TSLPF(LogPosteriorFunction):
         return self.data.fluxes
 
     @property
-    def time(self) -> list[ndarray]:
-        return self.data.time
+    def times(self) -> list[ndarray]:
+        return self.data.times
 
     @property
-    def wavelength(self) -> list[ndarray]:
-        return self.data.wavelength
+    def wavelengths(self) -> list[ndarray]:
+        return self.data.wavelengths
 
     @property
-    def ferr(self) -> list[ndarray]:
+    def errors(self) -> list[ndarray]:
         return self.data.errors
 
-    def set_data(self, data: TSDataSet):
+    def set_data(self, data: TSDataGroup):
         self._original_data = deepcopy(data)
         self.data = data
         self.npb: list[int] = [f.shape[0] for f in self.flux]
         self.npt: list[int] = [f.shape[1] for f in self.flux]
-        for i, time in enumerate(self.time):
+        for i, time in enumerate(self.times):
             self.tms[i].set_data(time)
-        if self._nm in (NM_GP_FIXED, NM_GP_FREE):
-            self._init_gp()
 
     def _init_parameters(self) -> None:
         self.ps = ParameterSet([])
         self._init_p_star()
         self._init_p_orbit()
-        self._init_limb_darkening()
+        self._init_p_transit_centers()
+        self._init_p_limb_darkening()
         self._init_p_radius_ratios()
+        self._init_p_noise()
+        if self._nm == NM_GP_FREE:
+            self._init_p_gp()
+        self._init_p_baseline()
+        self._init_p_bias()
         self.ps.freeze()
         self.ndim = len(self.ps)
 
@@ -192,6 +230,10 @@ class TSLPF(LogPosteriorFunction):
         self._nm = noise_models[noise_model]
         if self._nm in (NM_GP_FIXED, NM_GP_FREE):
             self._init_gp()
+        if self._nm == NM_GP_FREE:
+            self.ps.thaw()
+            self._init_p_gp()
+            self.ps.freeze()
 
     def _init_gp(self) -> None:
         """Initializes the Gaussian Process (GP) .
@@ -203,23 +245,26 @@ class TSLPF(LogPosteriorFunction):
         self._gp_ferr = []
         self._gp = []
         for d in self.data:
-            self._gp_time.append((tile(d.time[newaxis, :], (d.nwl, 1)) + arange(d.nwl)[:, newaxis]).ravel())
-            self._gp_flux.append(d.fluxes.ravel())
-            self._gp_ferr.append(d.errors.ravel())
+            self._gp_time.append((tile(d.time[newaxis, :], (d.nwl, 1)) + arange(d.nwl)[:, newaxis])[d.mask])
+            self._gp_flux.append(d.fluxes[d.mask])
+            self._gp_ferr.append(d.errors[d.mask])
             self._gp.append(GP(terms.Matern32Term(sigma=self._gp_flux[-1].std(), rho=0.1)))
             self._gp[-1].compute(self._gp_time[-1], yerr=self._gp_ferr[-1], quiet=True)
 
-    def set_gp_hyperparameters(self, sigma: float, rho: float) -> None:
+    def set_gp_hyperparameters(self, sigma: float, rho: float, idata: int | None = None) -> None:
         """Sets the Gaussian Process hyperparameters assuming a Matern32 kernel.
 
         Parameters
         ----------
-        sigma : float
+        sigma
             The kernel amplitude parameter.
 
-        rho : float
+        rho
             The length scale parameter.
 
+        idata
+            The data set for which to set the hyperparameters. If None, the hyperparameters are set for all data sets.
+
         Raises
         ------
         RuntimeError
@@ -227,9 +272,10 @@ class TSLPF(LogPosteriorFunction):
         """
         if self._gp is None:
             raise RuntimeError('The GP needs to be initialized before setting hyperparameters.')
-        for i, gp in enumerate(self._gp):
-            gp.kernel = terms.Matern32Term(sigma=sigma, rho=rho)
-            gp.compute(self._gp_time[i], yerr=self._gp_ferr[i], quiet=True)
+
+        for i in ([idata] if idata is not None else range(self.data.size)):
+            self._gp[i].kernel = terms.Matern32Term(sigma=sigma, rho=rho)
+            self._gp[i].compute(self._gp_time[i], yerr=self._gp_ferr[i], quiet=True)
 
     def set_gp_kernel(self, kernel: terms.Term) -> None:
         """Sets the kernel for the Gaussian Process (GP) model and recomputes the GP.
@@ -247,7 +293,7 @@ class TSLPF(LogPosteriorFunction):
         pstar = [GParameter('rho', 'stellar density', 'g/cm^3', UP(0.1, 25.0), (0, inf))]
         self.ps.add_global_block('star', pstar)
 
-    def _init_limb_darkening(self) -> None:
+    def _init_p_limb_darkening(self) -> None:
         if isinstance(self.ldmodel, LDTkLD):
             pld = [GParameter('teff', 'stellar TEff', 'K', NP(*self.ldmodel.sc.teff), (0, inf)),
                    GParameter('logg', 'stellar log g', '', NP(*self.ldmodel.sc.logg), (0, inf)),
@@ -263,8 +309,7 @@ class TSLPF(LogPosteriorFunction):
 
     def _init_p_orbit(self):
         ps = self.ps
-        pp = [GParameter('tc', 'zero_epoch', '', NP(0.0, 0.1), (-inf, inf)),
-              GParameter('p', 'period', 'd', NP(1.0, 1e-5), (0, inf)),
+        pp = [GParameter('p', 'period', 'd', NP(1.0, 1e-5), (0, inf)),
               GParameter('b', 'impact_parameter', 'R_s', UP(0.0, 1.0), (0, inf)),
               GParameter('secw', 'sqrt(e) cos(w)', '', NP(0.0, 1e-5), (-1, 1)),
               GParameter('sesw', 'sqrt(e) sin(w)', '', NP(0.0, 1e-5), (-1, 1))]
@@ -272,6 +317,14 @@ class TSLPF(LogPosteriorFunction):
         self._start_orbit = ps.blocks[-1].start
         self._sl_orbit = ps.blocks[-1].slice
 
+    def _init_p_transit_centers(self):
+        ps = self.ps
+        neps = max(self.data.epoch_groups) + 1
+        pp = [GParameter(f'tc_{i:02d}', f'zero_epoch_{i:02d}', '', NP(0.0, 0.1), (-inf, inf)) for i in range(neps)]
+        ps.add_global_block('transit_centers', pp)
+        self._start_tcs = ps.blocks[-1].start
+        self._sl_tcs = ps.blocks[-1].slice
+
     def _init_p_radius_ratios(self):
         ps = self.ps
         pp = [GParameter(f'k_{k:08.5f}', fr'radius ratio at {k:08.5f} $\mu$m', 'A_s', UP(0.02, 0.2), (0, inf)) for k in self.k_knots]
@@ -279,6 +332,50 @@ class TSLPF(LogPosteriorFunction):
         self._start_rratios = ps.blocks[-1].start
         self._sl_rratios = ps.blocks[-1].slice
 
+    def _init_p_noise(self):
+        ps = self.ps
+        pp = [GParameter(f'sigma_m_{i:02d}', f'Noise group {i} sigma multipler', '',
+                         NP(1.0, 0.01), (0, inf)) for i in range(self.data.n_noise_groups)]
+        ps.add_global_block('white_noise_multipliers', pp)
+        self._start_wnm = ps.blocks[-1].start
+        self._sl_wnm = ps.blocks[-1].slice
+
+    def _init_p_gp(self):
+        ps = self.ps
+        if not hasattr(self, '_sl_gp'):
+            pp = [GParameter('gp_log_sigma', 'GP log sigma', '', NP(0.0, 0.01), (-inf, inf)),
+                  GParameter('gp_log_rho', 'GP log rho', '', NP(0.0, 0.01), (-inf, inf))]
+            ps.add_global_block('gp_hyperparameters', pp)
+            self._start_gp = ps.blocks[-1].start
+            self._sl_gp = ps.blocks[-1].slice
+
+    def _init_p_baseline(self):
+        ps = self.ps
+        self.n_baselines = self.data.n_baselines
+        self.baseline_knots = []
+        pp = []
+        for i, d in enumerate(self.data):
+            if d.n_baseline== 1:
+                self.baseline_knots.append([])
+                pp.append(GParameter(f'bl_{i:02d}_c', 'baseline constant', '', NP(1.0, 1e-6), (0, inf)))
+            elif d.n_baseline > 1:
+                knots = linspace(d.wavelength.min(), d.wavelength.max(), d.n_baseline)
+                self.baseline_knots.append(knots)
+                pp.extend([GParameter(f'bl_{i:02d}_{k:08.5f}', fr'baseline at {k:08.5f} $\mu$m', '', NP(1.0, 1e-6), (0, inf)) for k in knots])
+        ps.add_global_block('baseline_coefficients', pp)
+        self._start_baseline = ps.blocks[-1].start
+        self._sl_baseline = ps.blocks[-1].slice
+
+    def _init_p_bias(self):
+        ps = self.ps
+        offset_groups = self.data.offset_groups
+        pp = []
+        for i in range(1, max(offset_groups) + 1):
+            pp.append(GParameter(f'bias_{i:02d}', 'bias offset', '', NP(0.0, 1e-6), (-inf, inf)))
+        ps.add_global_block('bias_offsets', pp)
+        self._start_bias = ps.blocks[-1].start
+        self._sl_bias = ps.blocks[-1].slice
+
     def set_ldtk_prior(self, teff, logg, metal, dataset: str = 'visir-lowres', width: float = 50,
                        uncertainty_multiplier: float = 10):
         hw = 0.5 * width
@@ -358,7 +455,7 @@ class TSLPF(LogPosteriorFunction):
             # Resample the radius ratios
             # --------------------------
             for i in range(den.shape[0]):
-                den[i, sln] = resample(xn, xo, deo[i, slo])
+                den[i, sln] = self._ip(xn, xo, deo[i, slo])
 
             self._de_population = den
             self.de = None
@@ -379,7 +476,7 @@ class TSLPF(LogPosteriorFunction):
             # Resample the radius ratios
             # --------------------------
             for i in range(fmcn.shape[0]):
-                fmcn[i, sln] = resample(xn, xo, fmco[i, slo])
+                fmcn[i, sln] = self._ip(xn, xo, fmco[i, slo])
 
             self._mc_chains = fmcn.reshape([mco.shape[0], mco.shape[1], ndn])
             self.sampler = None
@@ -429,7 +526,7 @@ class TSLPF(LogPosteriorFunction):
             # Resample the radius ratios
             # --------------------------
             for i in range(pvpn.shape[0]):
-                pvpn[i, sldn] = resample(xn, xo, pvpo[i, sldo])
+                pvpn[i, sldn] = self._ip(xn, xo, pvpo[i, sldo])
 
             self.de = None
             self._de_population = pvpn
@@ -451,9 +548,9 @@ class TSLPF(LogPosteriorFunction):
         """
         pvp = atleast_2d(pvp)
         ks = [zeros((pvp.shape[0], npb)) for npb in self.npb]
-        for ids in range(self.data.ngroups):
+        for ids in range(self.data.size):
             for ipv in range(pvp.shape[0]):
-                ks[ids][ipv,:] =  splev(self.wavelength[ids], splrep(self.k_knots, pvp[ipv], s=0.0))
+                ks[ids][ipv,:] =  self._ip(self.wavelengths[ids], self.k_knots, pvp[ipv])
         return ks
 
     def _eval_ldc(self, pvp):
@@ -467,10 +564,10 @@ class TSLPF(LogPosteriorFunction):
             pvp = atleast_2d(pvp)
             ldk = pvp[:, self._sl_ld].reshape([pvp.shape[0], self.nldc, 2])
             ldp = [zeros((pvp.shape[0], npb, 2)) for npb in self.npb]
-            for ids in range(self.data.ngroups):
+            for ids in range(self.data.size):
                 for ipv in range(pvp.shape[0]):
-                    ldp[ids][ipv, :, 0] = splev(self.wavelength[ids], splrep(self.ld_knots, ldk[ipv, :, 0], s=0.0))
-                    ldp[ids][ipv, :, 1] = splev(self.wavelength[ids], splrep(self.ld_knots, ldk[ipv, :, 1], s=0.0))
+                    ldp[ids][ipv, :, 0] = self._ip(self.wavelengths[ids], self.ld_knots, ldk[ipv, :, 0])
+                    ldp[ids][ipv, :, 1] = self._ip(self.wavelengths[ids], self.ld_knots, ldk[ipv, :, 1])
             return ldp
 
     def transit_model(self, pv, copy=True):
@@ -500,19 +597,56 @@ class TSLPF(LogPosteriorFunction):
         """
         pv = atleast_2d(pv)
         ldp = self._eval_ldc(pv)
-        t0 = pv[:, 1]
-        p = pv[:, 2]
+        t0s = pv[:, self._sl_tcs]
         k = self._eval_k(pv[:, self._sl_rratios])
+        p = pv[:, 1]
         aor = as_from_rhop(pv[:, 0], p)
-        inc = i_from_ba(pv[:, 3], aor)
-        ecc = pv[:, 4] ** 2 + pv[:, 5] ** 2
-        w = arctan2(pv[:, 5], pv[:, 4])
-        return [tm.evaluate(k[i], ldp[i], t0, p, aor, inc, ecc, w, copy) for i,tm in enumerate(self.tms)]
+        inc = i_from_ba(pv[:, 2], aor)
+        ecc = pv[:, 3] ** 2 + pv[:, 4] ** 2
+        w = arctan2(pv[:, 4], pv[:, 3])
+        epids = self.data.epoch_groups
+        fluxes = []
+        if isinstance(self.ldmodel, LDTkLD):
+            ldp, istar = self.ldmodel(self.tms[0].mu, ldp)
+            ldpi = dstack([ldp, istar])
+            for i, tm in enumerate(self.tms):
+                fluxes.append(tm.evaluate(k[i], ldpi[:, self.ldmodel.wlslices[i], :],
+                                          t0s[:, epids[i]], p, aor, inc, ecc, w, copy))
+        else:
+            for i, tm in enumerate(self.tms):
+                fluxes.append(tm.evaluate(k[i], ldp[i], t0s[:, epids[i]], p, aor, inc, ecc, w, copy))
+
+        for i, d in enumerate(self.data):
+            if d.offset_group > 0:
+                biases = pv[:, self._start_bias + d.offset_group - 1][:, None, None]
+                fluxes[i] = biases + (1.0 - biases) * fluxes[i]
+        return fluxes
+
+    def baseline_model(self, pv):
+        pv = atleast_2d(pv)[:, self._sl_baseline]
+        npv = pv.shape[0]
+        if self._baseline_models is None or self._baseline_models[0].shape[0] != npv:
+            self._baseline_models = [zeros((npv, d.nwl)) for d in self.data]
+        j = 0
+        for i, d in enumerate(self.data):
+            nbl = d.n_baseline
+            m = self._baseline_models[i]
+            if nbl == 1:
+                m[:, :] = pv[:, j][:, newaxis]
+            else:
+                for ipv in range(npv):
+                    m[ipv, :] = splev(d.wavelength, splrep(self.baseline_knots[i], pv[ipv, j:j+nbl], k=min(nbl-1, 3)))
+            j += nbl
+        return self._baseline_models
 
     def flux_model(self, pv):
-        return self.transit_model(pv)
+        transit_models = self.transit_model(pv)
+        baseline_models = self.baseline_model(pv)
+        for i in range(self.data.size):
+            transit_models[i][:, :, :] *= baseline_models[i][:, :, newaxis]
+        return transit_models
 
-    def create_pv_population(self, npop: int = 50):
+    def create_pv_population(self, npop: int = 50) -> ndarray:
         """ Crate a parameter vector population.
         Parameters
         ----------
@@ -530,7 +664,7 @@ class TSLPF(LogPosteriorFunction):
         for ipb in range(self.nk):
             self.set_prior(f'k_{ipb + 1:03d}', 'UP', kmin, kmax)
 
-    def lnlikelihood(self, pv):
+    def lnlikelihood(self, pv) -> ndarray | float :
         """Log likelihood for parameter vector pv.
 
         Parameters
@@ -547,16 +681,17 @@ class TSLPF(LogPosteriorFunction):
         pv = atleast_2d(pv)
         npv = pv.shape[0]
         fmod = self.flux_model(pv)
+        wn_multipliers = pv[:, self._sl_wnm]
         lnl = zeros(npv)
         if self._nm == NM_WHITE:
-            for i in range(self.data.ngroups):
-                lnl += lnlike_normal(self.flux[i], fmod[i], self.ferr[i])
-        elif self._nm == NM_GP_FIXED:
-            for j in range(npv):
-                for i in range(self.data.ngroups):
-                    lnl[j] += self._gp[i].log_likelihood(self._gp_flux[i] - fmod[i][j].ravel())
+            for i, d in enumerate(self.data):
+                lnl += lnlike_normal(d.fluxes, fmod[i], d.errors, wn_multipliers[:, d.noise_group], d.mask)
         else:
-            raise NotImplementedError("The free GP noise model hasn't been implemented yet.")
+            for j in range(npv):
+                if self._nm == NM_GP_FREE:
+                    self.set_gp_hyperparameters(*pv[j, self._sl_gp])
+                for i in range(self.data.size):
+                    lnl[j] += self._gp[i].log_likelihood(self._gp_flux[i] - fmod[i][j][self.data[i].mask])
         return lnl if npv > 1 else lnl[0]
 
     def create_initial_population(self, n: int, source: str, add_noise: bool = True) -> ndarray:
@@ -599,10 +734,10 @@ class TSLPF(LogPosteriorFunction):
             else:
                 pvp = rng.choice(pvs.reshape([-1, self.ndim]), size=n)
 
+        if pvp[0, self._sl_baseline][0] < 0.5:
+            pvp[:, self._sl_baseline] = rng.normal(1.0, 1e-6, size=(n, sum(self.n_baselines)))
+
         if add_noise:
-            pvp[:, 0] += rng.normal(0, 0.005, size=n)
-            pvp[:, 1] += rng.normal(0, 0.001, size=n)
-            pvp[:, 3] += rng.normal(0, 0.005, size=n)
             pvp[:, self._sl_rratios] += rng.normal(0, 1, pvp[:, self._sl_rratios].shape) * 0.002 * pvp[:, self._sl_rratios]
             pvp[:, self._sl_ld] += rng.normal(0, 1, pvp[:, self._sl_ld].shape) * 0.002 * pvp[:, self._sl_ld]
         return pvp
@@ -636,3 +771,6 @@ class TSLPF(LogPosteriorFunction):
         super().sample_mcmc(niter, thin, repeats, npop=npop, population=population, label=label, reset=reset,
                             leave=leave, save=save, use_tqdm=use_tqdm, pool=pool, lnpost=lnpost, vectorize=vectorize)
         self._mc_chains = self.sampler.chain.copy()
+
+    def save(self, overwrite: bool = True) -> None:
+        self._runner.save(overwrite=overwrite)
